@@ -2,38 +2,42 @@ from __future__ import annotations
 
 import multiprocessing
 import platform
+from dataclasses import asdict
 from functools import partial
-from typing import Literal, Sequence, Union
+from typing import Literal
 
 import astropy.units as u
 import healpy as hp
 import numpy as np
+import numpy.typing as npt
 import quadpy
 from astropy.coordinates import solar_system_ephemeris
 from astropy.time import Time
-from numpy.typing import NDArray
 
-from ._component import Component
-from ._decorators import validate_ang, validate_freq, validate_pixels
-from ._ephemeris import get_obs_and_earth_positions
+from ._constants import SPECIFIC_INTENSITY_UNITS
+from ._interp import interpolate_source_parameters
+from ._ipd_dens_funcs import PartialComputeDensityFunc, construct_density_partials
 from ._line_of_sight import get_line_of_sight_endpoints
-from ._source_functions import (
-    SPECIFIC_INTENSITY_UNITS,
+from ._sky_coords import DISTANCE_TO_JUPITER, get_obs_and_earth_positions
+from ._source_funcs import (
+    get_bandpass_integrated_blackbody_emission,
     get_blackbody_emission,
     get_dust_grain_temperature,
     get_phase_function,
     get_scattering_angle,
 )
+from ._types import FrequencyOrWavelength, Pixels, SkyAngles
 from ._unit_vectors import get_unit_vectors_from_ang, get_unit_vectors_from_pixels
-from .models import model_registry
+from ._validators import (
+    validate_and_normalize_weights,
+    validate_ang,
+    validate_frequency_in_model_range,
+    validate_pixels,
+)
+from .ipd_models import model_registry
 
-SYS_PROC_START_METHOD = "fork" if "windows" not in platform.system().lower() else None
-
-DISTANCE_TO_JUPITER = u.Quantity(5.2, u.AU)
-
-HEALPixIndicies = Union[int, Sequence[int], NDArray[np.integer]]
-SkyAngles = Union[u.Quantity[u.deg], u.Quantity[u.rad]]
-FrequencyOrWavelength = Union[u.Quantity[u.Hz], u.Quantity[u.m]]
+PLATFORM = platform.system().lower()
+SYS_PROC_START_METHOD = "fork" if "windows" not in PLATFORM else None
 
 
 class Zodipy:
@@ -46,14 +50,13 @@ class Zodipy:
     Attributes:
         model (str): Name of the interplanetary dust model to use in the simulations.
             Defaults to DIRBE.
-        extrapolate (bool): If True all spectral quantities in the selected model are
-            linearly extrapolated to the requested frequency/wavelength. If False, an
-            Exception will be raised on requested frequencies/wavelengths outside of the
-            valid model range. Default is False.
-        parallel (bool): If True, input pointing sequences will be split among all
+        extrapolate (bool): If `True` all spectral quantities in the selected model are
+            linearly extrapolated to the requested frequency/wavelength. If `False`, an
+            exception is raised on requested frequencies/wavelengths outside of the
+            valid model range. Default is `False`.
+        parallel (bool): If `True`, input pointing sequences will be split among all
             available cores on the machine, and the emission will be computed in parallel.
-            This is useful for large simulations. NOTE: This will require the user to guard
-            the executed code in a `if __name__ == "__main__"` block. Default is False.
+            This is useful for large pointing chunks. Default is `False`.
         n_proc (int): Number of cores to use when parallel computation. Defaults is None,
             which will use all available cores.
         solar_cut (u.Quantity[u.deg]): Cutoff angle from the sun in degrees. The emission
@@ -65,8 +68,8 @@ class Zodipy:
             argument to `None`. Defaults to 5 degrees.
         solar_cut_fill_value (float): Fill value for the masked solar cut pointing.
             Defaults to `np.nan`.
-        gauss_quad_order (int): Order of the Gaussian-Legendre quadrature used to evaluate
-            the brightness integral. Default is 50 points.
+        gauss_quad_degree (int): Order of the Gaussian-Legendre quadrature used to evaluate
+            the brightness integral. Default is 100 points.
         los_dist_cut (u.Quantity[u.AU]): Radial distance from the Sun at which all line of
             sights are truncated. Defaults to 5.2 AU which is the distance to Jupiter.
         ephemeris (str): Ephemeris used to compute the positions of the observer and the
@@ -84,7 +87,7 @@ class Zodipy:
         n_proc: int | None = None,
         solar_cut: u.Quantity[u.deg] | None = None,
         solar_cut_fill_value: float = np.nan,
-        gauss_quad_order: int = 100,
+        gauss_quad_degree: int = 100,
         los_dist_cut: u.Quantity[u.AU] = DISTANCE_TO_JUPITER,
         ephemeris: str = "de432s",
     ) -> None:
@@ -95,12 +98,12 @@ class Zodipy:
         self.n_proc = n_proc
         self.solar_cut = solar_cut.to(u.rad) if solar_cut is not None else solar_cut
         self.solar_cut_fill_value = solar_cut_fill_value
-        self.gauss_quad_order = gauss_quad_order
+        self.gauss_quad_degree = gauss_quad_degree
         self.los_dist_cut = los_dist_cut
         self.ephemeris = ephemeris
 
-        self._model = model_registry.get_model(model)
-        self._integration_scheme = quadpy.c1.gauss_legendre(self.gauss_quad_order)
+        self.ipd_model = model_registry.get_model(model)
+        self._integration_scheme = quadpy.c1.gauss_legendre(self.gauss_quad_degree)
 
     @property
     def supported_observers(self) -> list[str]:
@@ -108,8 +111,6 @@ class Zodipy:
 
         return list(solar_system_ephemeris.bodies) + ["semb-l2"]
 
-    @validate_freq
-    @validate_ang
     def get_emission_ang(
         self,
         freq: FrequencyOrWavelength,
@@ -118,6 +119,7 @@ class Zodipy:
         obs_time: Time,
         obs: str = "earth",
         obs_pos: u.Quantity[u.AU] | None = None,
+        weights: u.Quantity[u.MJy / u.sr] | None = None,
         lonlat: bool = False,
         return_comps: bool = False,
         coord_in: Literal["E", "G", "C"] = "E",
@@ -128,7 +130,8 @@ class Zodipy:
         the sky given by `theta` and `phi`.
 
         Args:
-            freq: Frequency or wavelength at which to evaluate the zodiacal emission.
+            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
+                a bandpass over which to evaluate the zodiacal emission.
             theta: Angular co-latitude coordinate of a point, or a sequence of points, on
                 the celestial sphere. Must be in the range [0, π] rad. Units must be either
                 radians or degrees.
@@ -141,6 +144,8 @@ class Zodipy:
                 `Zodipy` instance. Defaults to 'earth'.
             obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                 Overrides the `obs` argument. Default is None.
+            weights: Bandpass weights corresponding the the frequencies in `freq`. The
+                weights are assumed to be in units of MJy/sr.
             lonlat: If True, input angles (`theta`, `phi`) are assumed to be longitude and
                 latitude, otherwise, they are co-latitude and longitude.
             return_comps: If True, the emission is returned component-wise. Defaults to False.
@@ -152,6 +157,8 @@ class Zodipy:
 
         """
 
+        theta, phi = validate_ang(theta=theta, phi=phi, lonlat=lonlat)
+
         unique_angles, indicies = np.unique(
             np.asarray([theta, phi]), return_inverse=True, axis=1
         )
@@ -162,29 +169,26 @@ class Zodipy:
             lonlat=lonlat,
         )
 
-        emission = self._compute_emission(
+        return self._compute_emission(
             freq=freq,
+            weights=weights,
             obs=obs,
             obs_time=obs_time,
             obs_pos=obs_pos,
             unit_vectors=unit_vectors,
             indicies=indicies,
+            return_comps=return_comps,
         )
 
-        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)
-
-        return emission if return_comps else emission.sum(axis=0)
-
-    @validate_freq
-    @validate_pixels
     def get_emission_pix(
         self,
         freq: FrequencyOrWavelength,
-        pixels: HEALPixIndicies,
+        pixels: Pixels,
         nside: int,
         obs_time: Time,
         obs: str = "earth",
         obs_pos: u.Quantity[u.AU] | None = None,
+        weights: u.Quantity[u.MJy / u.sr] | None = None,
         return_comps: bool = False,
         coord_in: Literal["E", "G", "C"] = "E",
     ) -> u.Quantity[u.MJy / u.sr]:
@@ -194,7 +198,8 @@ class Zodipy:
         given by `nside`.
 
         Args:
-            freq: Frequency or wavelength at which to evaluate the zodiacal emission.
+            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
+                a bandpass over which to evaluate the zodiacal emission.
             pixels: HEALPix pixel indicies representing points on the celestial sphere.
             nside: HEALPix resolution parameter of the pixels and the binned map.
             obs_time: Time of observation.
@@ -203,6 +208,8 @@ class Zodipy:
                 `Zodipy` instance. Defaults to 'earth'.
             obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                 Overrides the `obs` argument. Default is None.
+            weights: Bandpass weights corresponding the the frequencies in `freq`. The
+                weights are assumed to be in units of MJy/sr.
             return_comps: If True, the emission is returned component-wise. Defaults to False.
             coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                 coordinates) by default.
@@ -212,6 +219,8 @@ class Zodipy:
 
         """
 
+        pixels = validate_pixels(pixels=pixels, nside=nside)
+
         unique_pixels, indicies = np.unique(pixels, return_inverse=True)
         unit_vectors = get_unit_vectors_from_pixels(
             coord_in=coord_in,
@@ -219,8 +228,9 @@ class Zodipy:
             nside=nside,
         )
 
-        emission = self._compute_emission(
+        return self._compute_emission(
             freq=freq,
+            weights=weights,
             obs=obs,
             obs_time=obs_time,
             obs_pos=obs_pos,
@@ -228,14 +238,9 @@ class Zodipy:
             indicies=indicies,
             pixels=unique_pixels,
             nside=nside,
+            return_comps=return_comps,
         )
 
-        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)
-
-        return emission if return_comps else emission.sum(axis=0)
-
-    @validate_ang
-    @validate_freq
     def get_binned_emission_ang(
         self,
         freq: FrequencyOrWavelength,
@@ -245,6 +250,7 @@ class Zodipy:
         obs_time: Time,
         obs: str = "earth",
         obs_pos: u.Quantity[u.AU] | None = None,
+        weights: u.Quantity[u.MJy / u.sr] | None = None,
         lonlat: bool = False,
         return_comps: bool = False,
         coord_in: Literal["E", "G", "C"] = "E",
@@ -256,7 +262,8 @@ class Zodipy:
         resolution given by `nside`.
 
         Args:
-            freq: Frequency or wavelength at which to evaluate the zodiacal emission.
+            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
+                a bandpass over which to evaluate the zodiacal emission.
             theta: Angular co-latitude coordinate of a point, or a sequence of points, on
                 the celestial sphere. Must be in the range [0, π] rad. Units must be either
                 radians or degrees.
@@ -270,6 +277,8 @@ class Zodipy:
                 `Zodipy` instance. Defaults to 'earth'.
             obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                 Overrides the `obs` argument. Default is None.
+            weights: Bandpass weights corresponding the the frequencies in `freq`. The
+                weights are assumed to be in units of MJy/sr.
             lonlat: If True, input angles `theta`, `phi` are assumed to be longitude and
                 latitude, otherwise, they are co-latitude and longitude.
             return_comps: If True, the emission is returned component-wise. Defaults to False.
@@ -281,6 +290,8 @@ class Zodipy:
 
         """
 
+        theta, phi = validate_ang(theta=theta, phi=phi, lonlat=lonlat)
+
         unique_angles, counts = np.unique(
             np.asarray([theta, phi]), return_counts=True, axis=1
         )
@@ -292,8 +303,9 @@ class Zodipy:
             lonlat=lonlat,
         )
 
-        emission = self._compute_emission(
+        return self._compute_emission(
             freq=freq,
+            weights=weights,
             obs=obs,
             obs_time=obs_time,
             obs_pos=obs_pos,
@@ -302,22 +314,18 @@ class Zodipy:
             binned=True,
             pixels=unique_pixels,
             nside=nside,
+            return_comps=return_comps,
         )
 
-        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)
-
-        return emission if return_comps else emission.sum(axis=0)
-
-    @validate_freq
-    @validate_pixels
     def get_binned_emission_pix(
         self,
         freq: FrequencyOrWavelength,
-        pixels: HEALPixIndicies,
+        pixels: Pixels,
         nside: int,
         obs_time: Time,
         obs: str = "earth",
         obs_pos: u.Quantity[u.AU] | None = None,
+        weights: u.Quantity[u.MJy / u.sr] | None = None,
         return_comps: bool = False,
         coord_in: Literal["E", "G", "C"] = "E",
     ) -> u.Quantity[u.MJy / u.sr]:
@@ -328,7 +336,8 @@ class Zodipy:
         `nside`.
 
         Args:
-            freq: Frequency or wavelength at which to evaluate the zodiacal emission.
+            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
+                a bandpass over which to evaluate the zodiacal emission.
             pixels: HEALPix pixel indicies representing points on the celestial sphere.
             nside: HEALPix resolution parameter of the pixels and the binned map.
             obs_time: Time of observation.
@@ -337,6 +346,8 @@ class Zodipy:
                 `Zodipy` instance. Defaults to 'earth'.
             obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                 Overrides the `obs` argument. Default is None.
+            weights: Bandpass weights corresponding the the frequencies in `freq`. The
+                weights are assumed to be in units of MJy/sr.
             return_comps: If True, the emission is returned component-wise. Defaults to False.
             coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                 coordinates) by default.
@@ -346,6 +357,8 @@ class Zodipy:
 
         """
 
+        pixels = validate_pixels(pixels=pixels, nside=nside)
+
         unique_pixels, counts = np.unique(pixels, return_counts=True)
         unit_vectors = get_unit_vectors_from_pixels(
             coord_in=coord_in,
@@ -353,8 +366,9 @@ class Zodipy:
             nside=nside,
         )
 
-        emission = self._compute_emission(
+        return self._compute_emission(
             freq=freq,
+            weights=weights,
             obs=obs,
             obs_time=obs_time,
             obs_pos=obs_pos,
@@ -363,41 +377,38 @@ class Zodipy:
             binned=True,
             pixels=unique_pixels,
             nside=nside,
+            return_comps=return_comps,
         )
 
-        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)
-
-        return emission if return_comps else emission.sum(axis=0)
-
     def _compute_emission(
         self,
-        freq: u.Quantity[u.GHz],
+        freq: FrequencyOrWavelength,
+        weights: u.Quantity[u.MJy / u.sr] | None,
         obs: str,
         obs_time: Time,
-        unit_vectors: NDArray[np.floating],
-        indicies: NDArray[np.integer],
+        unit_vectors: npt.NDArray[np.float64],
+        indicies: npt.NDArray[np.int64],
         binned: bool = False,
         obs_pos: u.Quantity[u.AU] | None = None,
-        pixels: NDArray[np.integer] | None = None,
+        pixels: npt.NDArray[np.int64] | None = None,
         nside: int | None = None,
-    ) -> NDArray[np.floating]:
+        return_comps: bool = False,
+    ) -> u.Quantity[u.MJy / u.sr]:
         """Computes the component-wise zodiacal emission."""
 
+        if not self.extrapolate:
+            validate_frequency_in_model_range(freq=freq, model=self.ipd_model)
+
+        normalized_weights = validate_and_normalize_weights(freq=freq, weights=weights)
+
+        interpolated_source_params = interpolate_source_parameters(
+            model=self.ipd_model, freq=freq, weights=normalized_weights
+        )
+
         observer_position, earth_position = get_obs_and_earth_positions(
             obs=obs, obs_time=obs_time, obs_pos=obs_pos
         )
 
-        if self._model.solar_irradiance_model is not None:
-            solar_irradiance = (
-                self._model.solar_irradiance_model.interpolate_solar_irradiance(
-                    freq=freq,
-                    albedos=self._model.albedos,
-                    extrapolate=self.extrapolate,
-                )
-            )
-        else:
-            solar_irradiance = 0
-
         start, stop = get_line_of_sight_endpoints(
             cutoff=self.los_dist_cut.value,
             obs_pos=observer_position,
@@ -405,17 +416,46 @@ class Zodipy:
         )
 
         emission = np.zeros(
-            (self._model.n_comps, hp.nside2npix(nside) if binned else indicies.size)
+            (self.ipd_model.n_comps, hp.nside2npix(nside) if binned else indicies.size)
         )
 
-        # Preparing quantities for broadcasting
-        observer_position_expanded = np.expand_dims(observer_position, axis=-1)
-        earth_position_expanded = np.expand_dims(earth_position, axis=-1)
+        # Convert to Hz if `freq` is in units of wavelength
+        if not freq.unit.is_equivalent(u.Hz):
+            freq = freq.to(u.Hz, u.spectral())
 
-        emissivities, albedos, phase_coefficients = self._model.interp_source_params(
-            freq
+            # Flip weights if `freq` is in units of wavelength
+            if normalized_weights is not None:
+                normalized_weights = np.flip(normalized_weights) / np.trapz(
+                    normalized_weights, freq.value
+                )
+
+        # Some components require additional non-static parameters to be computed, such as
+        # the earth-trailing feature which needs the earth position in addition to the
+        # model parameters.
+        computed_parameters = {
+            "X_earth": np.expand_dims(earth_position, axis=-1),
+        }
+        # Construct component density functions
+        density_partials = construct_density_partials(
+            comps=list(self.ipd_model.comps.values()),
+            computed_params=computed_parameters,
         )
-        # Distribute pointing to available CPUs and compute the emission in parallel.
+        # Create partial function with fixed arguments. This partial will again be used
+        # as the basis for another partial depnding on wether or not the code is run in
+        # parallel.
+        partial_common_integrand = partial(
+            _get_emission_at_step,
+            start=start,
+            gauss_quad_degree=self.gauss_quad_degree,
+            X_obs=np.expand_dims(observer_position, axis=-1),
+            density_partials=density_partials,
+            freq=freq.value,
+            weights=normalized_weights,
+            T_0=self.ipd_model.T_0,
+            delta=self.ipd_model.delta,
+            **asdict(interpolated_source_params),
+        )
+        # Distribute pointing to CPUs and compute the emission in parallel.
         if self.parallel:
             n_proc = multiprocessing.cpu_count() if self.n_proc is None else self.n_proc
 
@@ -431,21 +471,9 @@ class Zodipy:
                 # pointings, and p is the number of integration points of the quadrature.
                 partial_integrand_chunks = [
                     partial(
-                        _compute_emission_at_step,
-                        start=start,
+                        partial_common_integrand,
                         stop=np.expand_dims(stop, axis=-1),
-                        n_quad_points=self.gauss_quad_order,
-                        X_obs=observer_position_expanded,
-                        X_earth=earth_position_expanded,
                         u_los=np.expand_dims(unit_vectors, axis=-1),
-                        comps=list(self._model.comps.values()),
-                        freq=freq.value,
-                        T_0=self._model.T_0,
-                        delta=self._model.delta,
-                        emissivities=emissivities,
-                        albedos=albedos,
-                        phase_coefficients=phase_coefficients,
-                        solar_irradiance=solar_irradiance,
                     )
                     for unit_vectors, stop in zip(unit_vector_chunks, stop_chunks)
                 ]
@@ -466,26 +494,14 @@ class Zodipy:
             unit_vectors_expanded = np.expand_dims(unit_vectors, axis=-1)
             stop_expanded = np.expand_dims(stop, axis=-1)
 
-            partial_emission_integrand = partial(
-                _compute_emission_at_step,
-                start=start,
+            partial_integrand = partial(
+                partial_common_integrand,
                 stop=stop_expanded,
-                n_quad_points=self.gauss_quad_order,
-                X_obs=observer_position_expanded,
-                X_earth=earth_position_expanded,
                 u_los=unit_vectors_expanded,
-                comps=list(self._model.comps.values()),
-                freq=freq.value,
-                T_0=self._model.T_0,
-                delta=self._model.delta,
-                emissivities=emissivities,
-                albedos=albedos,
-                phase_coefficients=phase_coefficients,
-                solar_irradiance=solar_irradiance,
             )
 
             integrated_comp_emission = self._integration_scheme.integrate(
-                partial_emission_integrand, [-1, 1]
+                partial_integrand, [-1, 1]
             )
 
         # We convert the integral from [-1, 1] back to [start, stop].
@@ -510,7 +526,9 @@ class Zodipy:
             else:
                 emission[:, solar_mask[indicies]] = self.solar_cut_fill_value
 
-        return emission
+        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)
+
+        return emission if return_comps else emission.sum(axis=0)
 
     def __repr__(self) -> str:
         repr_str = f"{self.__class__.__name__}("
@@ -522,24 +540,23 @@ class Zodipy:
         return repr_str[:-2] + ")"
 
 
-def _compute_emission_at_step(
-    r: float | NDArray[np.floating],
-    *,
+def _get_emission_at_step(
+    r: float | npt.NDArray[np.float64],
     start: float,
-    stop: float | NDArray[np.floating],
-    n_quad_points: int,
-    X_obs: NDArray[np.floating],
-    X_earth: NDArray[np.floating],
-    u_los: NDArray[np.floating],
-    comps: list[Component],
-    freq: float,
+    stop: float | npt.NDArray[np.float64],
+    gauss_quad_degree: int,
+    X_obs: npt.NDArray[np.float64],
+    u_los: npt.NDArray[np.float64],
+    density_partials: tuple[PartialComputeDensityFunc],
+    freq: float | npt.NDArray[np.float64],
+    weights: npt.NDArray[np.float64] | None,
     T_0: float,
     delta: float,
-    emissivities: list[float],
-    albedos: list[float],
-    phase_coefficients: tuple[float, float, float],
+    emissivities: npt.NDArray[np.float64],
+    albedos: npt.NDArray[np.float64],
+    phase_coefficients: tuple[float, ...],
     solar_irradiance: float,
-) -> NDArray[np.floating]:
+) -> npt.NDArray[np.float64]:
     """Returns the zodiacal emission at a step along all lines of sight."""
 
     # Convert the line of sight range from [-1, 1] to the true ecliptic positions
@@ -550,13 +567,24 @@ def _compute_emission_at_step(
     R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)
 
     temperature = get_dust_grain_temperature(R_helio, T_0, delta)
-    blackbody_emission = get_blackbody_emission(freq, temperature)
 
-    emission = np.zeros((len(comps), np.shape(X_helio)[1], n_quad_points))
-    density = np.zeros((len(comps), np.shape(X_helio)[1], n_quad_points))
+    if weights is not None:
+        blackbody_emission = get_bandpass_integrated_blackbody_emission(
+            freq=freq,
+            weights=weights,
+            T=temperature,
+        )
+    else:
+        blackbody_emission = get_blackbody_emission(freq=freq, T=temperature)
 
-    for idx, (comp, albedo, emissivity) in enumerate(zip(comps, albedos, emissivities)):
-        density[idx] = comp.compute_density(X_helio, X_earth=X_earth)
+    emission = np.zeros(
+        (len(density_partials), np.shape(X_helio)[1], gauss_quad_degree)
+    )
+    density = np.zeros_like(emission)
+    for idx, (get_density_func, albedo, emissivity) in enumerate(
+        zip(density_partials, albedos, emissivities)
+    ):
+        density[idx] = get_density_func(X_helio)
         emission[idx] = (1 - albedo) * (emissivity * blackbody_emission)
 
     if any(albedo != 0 for albedo in albedos):
